using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Rendering;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace CheesyFX
{
    public class HeadLight
    {
        private MVRScript script;
        private int id;
        public string name;
        
        public GameObject gameObject;
        private Light light;

        public JSONStorableBool enabled;
        
        private JSONStorableStringChooser type;
        private JSONStorableStringChooser renderMode;
        private JSONStorableStringChooser shadowResolution;
        public JSONStorableFloat intensity;
        
        private JSONStorableFloat spotAngle;
        private JSONStorableFloat range;
        public JSONStorableBool castShadows;
        private JSONStorableFloat shadowStrength;
        public JSONStorableColor color;
        
        
        private JSONStorableFloat offsetX;
        private JSONStorableFloat rotation;
        
        private Button button;
        private RectTransform buttonTransform;
        private string buttonText;
        private ColorBlock activeColor;

        public HeadLight(MVRScript script, int id)
        {
            this.script = script;
            this.id = id;
            
            float offsetMin, offsetMax, offsetDefault, rotationMin, rotationMax, rotationDefault, intensityDefault;
            switch (id)
            {
                case 0:
                    offsetMin = -3f;
                    offsetMax = 0f;
                    offsetDefault = -.75f;
                    rotationDefault = 30f;
                    rotationMin = 0f;
                    rotationMax = 90f;
                    intensityDefault = .5f;
                    buttonText = "L";
                    name = "Left";
                    break;
                case 1:
                    offsetMin = -3f;
                    offsetMax = 3f;
                    offsetDefault = 0f;
                    rotationDefault = 0f;
                    rotationMin = -90f;
                    rotationMax = 90f;
                    intensityDefault = .3f;
                    buttonText = "C";
                    name = "Center";
                    break;
                default:
                    offsetMin = 0f;
                    offsetMax = 3f;
                    offsetDefault = .75f;
                    rotationDefault = -30f;
                    rotationMin = -90f;
                    rotationMax = 0f;
                    intensityDefault = .5f;
                    buttonText = "R";
                    name = "Right";
                    break;
            }
            Vector3 offset = new Vector3(offsetDefault, 0f, 0f);
            float spotAngleDefault = LightMeUp.isOnVR ? 120f : 90f;
            
            gameObject = new GameObject();
            Transform transform = gameObject.transform;
            transform.parent = LightMeUp.cameraTransform;
            transform.localPosition = new Vector3();
            transform.localRotation = Quaternion.identity;
            transform.localPosition = offset;
            
            light = gameObject.AddComponent<Light>();

            enabled = new JSONStorableBool("Enabled", true, EnabledCallback);
            type = new JSONStorableStringChooser("Type", Enum.GetNames(typeof(LightType)).ToList(), LightType.Spot.ToString(), "Type", TypeCallback);
            renderMode = new JSONStorableStringChooser("Render Mode", Enum.GetNames(typeof(LightRenderMode)).ToList(), LightRenderMode.Auto.ToString(), "Render Mode", RenderModeCallback);
            shadowResolution = new JSONStorableStringChooser("Shadow Resolution", Enum.GetNames(typeof(LightShadowResolution)).ToList(), LightShadowResolution.Low.ToString(), "Shadow Resolution",ShadowResolutionCallback);
            intensity = new JSONStorableFloat("Intensity", intensityDefault, IntensityCallback, 0f, 2f, true);
            range = new JSONStorableFloat("Range", 5f, RangeCallback, 0f, 10f, false);
            shadowStrength = new JSONStorableFloat("Shadow Strength", .2f, ShadowStrengthCallback, 0f, 1f, false);
            spotAngle = new JSONStorableFloat("Spot Angle", spotAngleDefault, SpotAngleCallback, 0f, 180f, false);
            castShadows = new JSONStorableBool("Cast Shadows", true, CastShadowsCallback);
            color = new JSONStorableColor("Color", LightMeUp.defaultColors[id].ToHSV(), ColorCallback);
            offsetX = new JSONStorableFloat("Horizontal Offset",offset.x, OffsetCallback, offsetMin, offsetMax);
            rotation = new JSONStorableFloat("Rotation",rotationDefault, RotationCallback, rotationMin, rotationMax);
            
            light.color = LightMeUp.defaultColors[id];
            range.val = 5f;
            light.intensity = intensity.defaultVal;
            light.spotAngle = spotAngle.defaultVal;
            light.type = LightType.Spot;
            gameObject.transform.localRotation = Quaternion.Euler(0f, rotationDefault, 0f);
            light.shadows = LightShadows.Hard;
            light.shadowResolution = LightShadowResolution.Low;
            light.shadowStrength = .2f;
        }

        public void Toggle()
        {
            enabled.val = !enabled.val;
        }

        public void EnabledCallback(bool b)
        {
            gameObject.SetActive(b);
            button.colors = b ? activeColor : LightMeUp.disabledColorBlock;
        }

        private void CastShadowsCallback(bool b)
        {
            light.shadows = b ? LightShadows.Hard : LightShadows.None;
        }

        private void ShadowStrengthCallback(float f)
        {
            light.shadowStrength = f;
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.shadowStrength = f;
                headLight.shadowStrength.val = f;
            }
        }
        
        private void ShadowResolutionCallback(string x)
        {
            light.shadowResolution = (LightShadowResolution)Enum.Parse(typeof(LightShadowResolution), x);
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.shadowResolution = (LightShadowResolution)Enum.Parse(typeof(LightShadowResolution), x);
                headLight.type.valNoCallback = x;
            }
        }

        private void TypeCallback(string x)
        {
            light.type = (LightType)Enum.Parse(typeof(LightType), x);
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.type = (LightType)Enum.Parse(typeof(LightType), x);
                headLight.type.valNoCallback = x;
            }
        }

        private void RenderModeCallback(string x)
        {
            light.renderMode = (LightRenderMode)Enum.Parse(typeof(LightRenderMode), x);
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.renderMode = (LightRenderMode)Enum.Parse(typeof(LightRenderMode), x);
                headLight.type.valNoCallback = x;
            }
        }

        private void IntensityCallback(float f)
        {
            light.intensity = f;
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.intensity = f;
                headLight.intensity.valNoCallback = f;
            }
        }

        private void RangeCallback(float f)
        {
            light.range = f;
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.range = f;
                headLight.range.valNoCallback = f;
            }
        }

        private void SpotAngleCallback(float f)
        {
            light.spotAngle = f;
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.spotAngle = f;
                headLight.spotAngle.valNoCallback = f;
            }
        }

        private void OffsetCallback(float f)
        {
            gameObject.transform.localPosition = new Vector3(f, 0f, 0f);
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.gameObject.transform.localPosition = new Vector3(-f, 0f, 0f);
                headLight.offsetX.valNoCallback = -f;
            }
        }

        private void RotationCallback(float r)
        {
            Quaternion rotation = Quaternion.Euler(0f, r, 0f);
            gameObject.transform.localRotation = rotation;
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                rotation = Quaternion.Euler(0f, -r, 0f);
                headLight.gameObject.transform.localRotation = rotation;
                headLight.rotation.valNoCallback = -r;
            } 
        }

        private void ColorCallback(float h, float s, float v)
        {
            Color color = Color.HSVToRGB(h, s, v);
            light.color = color;
            activeColor.normalColor = color;
            activeColor.pressedColor = color;
            activeColor.highlightedColor = color;
            if (enabled.val) button.colors = activeColor;
        }

        public void UpdateRotation(Vector3 focusPoint)
        {
            Quaternion goal = new Quaternion();
            goal.SetLookRotation(focusPoint - gameObject.transform.position);
            Quaternion current = gameObject.transform.rotation;
            gameObject.transform.rotation = Quaternion.Lerp(current, goal, 4*Time.deltaTime);
            rotation.valNoCallback = gameObject.transform.localRotation.eulerAngles[1];
        }

        public void Destroy()
        {
            Object.Destroy(gameObject);
            Object.Destroy(buttonTransform.gameObject);
        }

        public List<object> CreateUI()
        {
            List<object> UIElements = new List<object>
            {
                script.CreatePopup(type),
                script.CreatePopup(renderMode),
                script.CreateSlider(intensity, false),
                script.CreateSlider(range),
                script.CreateSlider(spotAngle),
                script.CreateColorPicker(color, true),
                CreateSpacer(10f, true),
                script.CreateSlider(rotation, true),
                script.CreateSlider(offsetX, true),
                script.CreateSlider(shadowStrength, false),
                script.CreateScrollablePopup(shadowResolution, true),
                script.CreateToggle(castShadows)
            };
            return UIElements;
        }

        private UIDynamic CreateSpacer(float hight, bool rightside)
        {
            UIDynamic spacer = script.CreateSpacer(rightside);
            spacer.height = hight;
            return spacer;
        }

        public void CreateHUDToggle(RectTransform parent)
        {
            buttonTransform = script.manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
            buttonTransform = Object.Instantiate(buttonTransform, parent);
            buttonTransform.name = $"HeadLight{id}";
            buttonTransform.anchorMax = new Vector2(.25f+.13f*(id+1)/3, 1);
            buttonTransform.anchorMin = new Vector2(.25f+.13f*id/3, 0);
            buttonTransform.offsetMax = new Vector2(0, -5f);
            buttonTransform.offsetMin = new Vector2(0, 5f);
            button = buttonTransform.GetComponent<Button>();
            buttonTransform.Find("Text").GetComponent<Text>().text = this.buttonText;
            
            button.onClick.AddListener(Toggle);
            activeColor = ColorBlock.defaultColorBlock;
            var color1 = light.color;
            activeColor.normalColor = color1;
            activeColor.pressedColor = color1;
            activeColor.highlightedColor = color1;
            button.colors = activeColor;
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            enabled.Store(jc);
            color.Store(jc);
            type.Store(jc);
            renderMode.Store(jc);
            shadowResolution.Store(jc);
            intensity.Store(jc);
            spotAngle.Store(jc);
            range.Store(jc);
            castShadows.Store(jc);
            shadowStrength.Store(jc);
            offsetX.Store(jc);
            rotation.Store(jc);
            return jc;
        }

        public void LoadJSON(JSONClass jc)
        {
            enabled.Load(jc);
            color.Load(jc);
            type.Load(jc);
            renderMode.Load(jc);
            shadowResolution.Load(jc);
            intensity.Load(jc);
            spotAngle.Load(jc);
            range.Load(jc);
            castShadows.Load(jc);
            shadowStrength.Load(jc);
            offsetX.Load(jc);
            rotation.Load(jc);
        }
    }

    internal static class ColorExt
    {
        public static HSVColor ToHSV(this Color color)
        {
            float h, s, v;
            Color.RGBToHSV(color, out h, out s, out v);
            return new HSVColor
            {
                H = h,
                S = s,
                V = v
            };
        }
    }

    internal static class ObjEct
    {
        public static void Print(this object obj)
        {
            SuperController.LogMessage(obj.ToString());
        }

        public static void NullCheck(this object obj)
        {
            (obj == null).Print();
        }
    }
    
    public static class IEumeratorExt
    {
        public static IEnumerator Start(this IEnumerator ienum)
        {
            if (ienum != null) SuperController.singleton.StartCoroutine(ienum);
            return ienum;
        }

        public static void Stop(this IEnumerator ienum)
        {
            if (ienum != null) SuperController.singleton.StopCoroutine(ienum);
        }
    }
}